import uuid

from requests import HTTPError

from pulpcore.client.pulp_certguard import CertguardRHSMCertGuard, ContentguardsRhsmApi

from pulp_certguard.tests.functional.api.base import BaseCertGuard, CommonDenialTestsMixin
from pulp_certguard.tests.functional.constants import (
    RHSM_CA_CERT_FILE_PATH,
    RHSM_CLIENT_CERT_FROM_UNTRUSTED_CA,
    RHSM_CLIENT_CERT_TRUSTED_BUT_EXPIRED,
    THIRDPARTY_CA_CERT_FILE_PATH,
    RHSM_UBER_CERT_BASE_PATH_ONE,
    RHSM_UBER_CERT_BASE_PATH_TWO,
    RHSM_UBER_CLIENT_CERT,
    RHSM_V1_ONE_AND_TWO_VAR_CLIENT_CERT,
    RHSM_V1_ONE_VAR_BASE_PATH,
    RHSM_V1_TWO_VAR_BASE_PATH,
    RHSM_V1_ZERO_VAR_CLIENT_CERT,
    RHSM_V1_ZERO_VAR_BASE_PATH,
    RHSM_V3_INVALID_BASE_PATH,
    RHSM_V3_ONE_AND_TWO_VAR_CLIENT_CERT,
    RHSM_V3_ONE_VAR_BASE_PATH,
    RHSM_V3_TWO_VAR_BASE_PATH,
    RHSM_V3_ZERO_VAR_CLIENT_CERT,
    RHSM_V3_ZERO_VAR_BASE_PATH,
)
from pulp_certguard.tests.functional.utils import (
    gen_certguard_client,
    set_distribution_base_path_and_download_a_content_unit_with_cert,
)


class RHSMCertGuardBase(BaseCertGuard):
    """A base class for all RHSMCertGuard tests which provides generic setup and teardown."""

    @classmethod
    def _setup_content_guard(cls):
        # Create a RHSM Content Guard
        certguard_client = gen_certguard_client()
        cls.rhsm_content_guards_api = ContentguardsRhsmApi(certguard_client)

        rhsm_ca_cert_data = cls._load_rhsm_ca_cert_file()

        rhsm_cert_guard = CertguardRHSMCertGuard(
            name=str(uuid.uuid4()),
            ca_certificate=rhsm_ca_cert_data
        )
        cls.rhsm_content_guard_data = cls.rhsm_content_guards_api.create(rhsm_cert_guard)
        cls.teardown_cleanups.append(
            (cls.rhsm_content_guards_api.delete, cls.rhsm_content_guard_data.pulp_href)
        )
        return cls.rhsm_content_guard_data.pulp_href

    @classmethod
    def _load_rhsm_ca_cert_file(cls):
        with open(RHSM_CA_CERT_FILE_PATH, 'r') as rhsm_ca_cert_data_file:
            rhsm_ca_cert_data = rhsm_ca_cert_data_file.read()

        return rhsm_ca_cert_data


class RHSMCABundleCertGuardBase(RHSMCertGuardBase):
    """A base class for all RHSMCertGuard tests with a CA-bundle file."""

    @classmethod
    def _load_rhsm_ca_cert_file(cls):
        with open(RHSM_CA_CERT_FILE_PATH, 'r') as rhsm_ca_cert_data_file:
            rhsm_ca_cert_data = rhsm_ca_cert_data_file.read()

        with open(THIRDPARTY_CA_CERT_FILE_PATH, 'r') as thirdparty_ca_cert_file:
            thirdparty_ca_cert_data = thirdparty_ca_cert_file.read()

        # build a bundle to use (i.e., a list-of-CA-certs we will trust)
        return thirdparty_ca_cert_data + rhsm_ca_cert_data


class RHSMV3CertGuardTestCase(RHSMCertGuardBase):
    """Api tests for RHSMCertGard with V3 RHSM Certificates."""

    def test_allow_request_when_cert_matches_zero_var_path(self):
        """
        Assert a correctly configured client can fetch content from a zero-variable path.

        1. Configure the distribution with a zero-variable path in the RHSM Cert.
        2. Attempt to download content.
        """
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            RHSM_V3_ZERO_VAR_BASE_PATH,
            self.repo.pulp_href,
            RHSM_V3_ZERO_VAR_CLIENT_CERT
        )

    def test_allow_request_when_cert_matches_one_var_path(self):
        """
        Assert a correctly configured client can fetch content from a one-variable path.

        1. Configure the distribution with a one-variable path in the RHSM Cert.
        2. Attempt to download content.
        """
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            RHSM_V3_ONE_VAR_BASE_PATH,
            self.repo.pulp_href,
            RHSM_V3_ONE_AND_TWO_VAR_CLIENT_CERT
        )

    def test_allow_request_when_cert_matches_two_var_path(self):
        """
        Assert a correctly configured client can fetch content from a two-variable path.

        1. Configure the distribution with a two-variable path in the RHSM Cert.
        2. Attempt to download content.
        """
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            RHSM_V3_TWO_VAR_BASE_PATH,
            self.repo.pulp_href,
            RHSM_V3_ONE_AND_TWO_VAR_CLIENT_CERT
        )

    def test_allow_request_when_requesting_the_distribution_root(self):
        """
        Assert a correctly configured client can fetch content from the root of a distribution.

        1. Configure the distribution with a zero-variable path in the RHSM Cert.
        2. Attempt to fetch the url of the distribution itself (its root).
        """
        content_path = ""  # This causes the root to be fetched
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            RHSM_V3_ZERO_VAR_BASE_PATH,
            self.repo.pulp_href,
            RHSM_V3_ZERO_VAR_CLIENT_CERT,
            content_path,
        )

    def test_allow_request_to_subdir_of_path(self):
        """
        Assert a correctly configured client can fetch content from a subdir of a distribution.

        1. Configure the distribution with a zero-variable path in the RHSM Cert.
        2. Attempt to download a content url with a subdir in it.
        3. Assert a 404 was received.
        """
        content_path = "somedir/made_up_content.iso"
        with self.assertRaises(HTTPError) as raised_exception:
            set_distribution_base_path_and_download_a_content_unit_with_cert(
                self.distribution.pulp_href,
                RHSM_V3_ZERO_VAR_BASE_PATH,
                self.repo.pulp_href,
                RHSM_V3_ZERO_VAR_CLIENT_CERT,
                content_path,
            )

        # The path doesn't exist so we expect a 404, but the authorization part we are testing works
        self.assertEqual(raised_exception.exception.response.status_code, 404)


class RHSMV1CertGuardTestCase(RHSMCertGuardBase):
    """Api tests for RHSMCertGard with V1 RHSM Certificates."""

    def test_allow_request_when_cert_matches_zero_var_path(self):
        """
        Assert a correctly configured client can fetch content from a zero-variable path.

        1. Configure the distribution with a zero-variable path in the RHSM Cert.
        2. Attempt to download content.
        """
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            RHSM_V1_ZERO_VAR_BASE_PATH,
            self.repo.pulp_href,
            RHSM_V1_ZERO_VAR_CLIENT_CERT
        )

    def test_allow_request_when_cert_matches_one_var_path(self):
        """
        Assert a correctly configured client can fetch content from a one-variable path.

        1. Configure the distribution with a one-variable path in the RHSM Cert.
        2. Attempt to download content.
        """
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            RHSM_V1_ONE_VAR_BASE_PATH,
            self.repo.pulp_href,
            RHSM_V1_ONE_AND_TWO_VAR_CLIENT_CERT
        )

    def test_allow_request_when_cert_matches_two_var_path(self):
        """
        Assert a correctly configured client can fetch content from a two-variable path.

        1. Configure the distribution with a two-variable path in the RHSM Cert.
        2. Attempt to download content.
        """
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            RHSM_V1_TWO_VAR_BASE_PATH,
            self.repo.pulp_href,
            RHSM_V1_ONE_AND_TWO_VAR_CLIENT_CERT
        )

    def test_allow_request_when_requesting_the_distribution_root(self):
        """
        Assert a correctly configured client can fetch content from the root of a distribution.

        1. Configure the distribution with a zero-variable path in the RHSM Cert.
        2. Attempt to fetch the url of the distribution itself (its root).
        """
        content_path = ""  # This causes the root to be fetched
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            RHSM_V1_ZERO_VAR_BASE_PATH,
            self.repo.pulp_href,
            RHSM_V1_ZERO_VAR_CLIENT_CERT,
            content_path,
        )

    def test_allow_request_to_subdir_of_path(self):
        """
        Assert a correctly configured client can fetch content from a subdir of a distribution.

        1. Configure the distribution with a zero-variable path in the RHSM Cert.
        2. Attempt to download a content url with a subdir in it.
        3. Assert a 404 was received.
        """
        content_path = "somedir/made_up_content.iso"
        with self.assertRaises(HTTPError) as raised_exception:
            set_distribution_base_path_and_download_a_content_unit_with_cert(
                self.distribution.pulp_href,
                RHSM_V1_ZERO_VAR_BASE_PATH,
                self.repo.pulp_href,
                RHSM_V1_ZERO_VAR_CLIENT_CERT,
                content_path,
            )

        # The path doesn't exist so we expect a 404, but the authorization part we are testing works
        self.assertEqual(raised_exception.exception.response.status_code, 404)


class RHSMUberCertTestCase(RHSMCertGuardBase):
    """Api tests for RHSMCertGard with an "Uber" Certificate."""

    def test_allow_request_with_uber_cert_for_any_subpath(self):
        """
        Assert a client with an uber cert can fetch any subpath.

        1. Configure the distribution with a subpath of the uber cert.
        2. Attempt to download content.
        3. Configure the distribution with a different subpath of the uber cert.
        4. Attempt to download content.
        """
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            RHSM_UBER_CERT_BASE_PATH_ONE,
            self.repo.pulp_href,
            RHSM_UBER_CLIENT_CERT
        )
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            RHSM_UBER_CERT_BASE_PATH_TWO,
            self.repo.pulp_href,
            RHSM_UBER_CLIENT_CERT
        )


class RHSMCertGuardDenialTestCase(RHSMCertGuardBase, CommonDenialTestsMixin):
    """Api tests for RHSMCertGard to assert denials for authorization."""

    DENIALS_BASE_PATH = RHSM_V3_ZERO_VAR_BASE_PATH
    UNTRUSTED_CLIENT_CERT_PATH = RHSM_CLIENT_CERT_FROM_UNTRUSTED_CA

    def test_denial_when_client_cert_does_not_contain_subpath_of_distribution_base_path(self):
        """
        Assert denial when a client with a cert that does not contain a subpath of the distribution.

        1. Configure the distribution with path that is not a subpath contained in the cert.
        2. Attempt to download content.
        3. Assert a 403 Unauthorized is returned.
        """
        with self.assertRaises(HTTPError) as raised_exception:
            set_distribution_base_path_and_download_a_content_unit_with_cert(
                self.distribution.pulp_href,
                RHSM_V3_INVALID_BASE_PATH,
                self.repo.pulp_href,
                RHSM_V3_ZERO_VAR_CLIENT_CERT
            )
        self.assertEqual(raised_exception.exception.response.status_code, 403)

    def test_denial_when_client_cert_is_trusted_but_expired(self):
        """
        Assert denial when a client sends a trusted but expired cert that has a valid subpath.

        1. Configure the distribution with valid path contained in the cert.
        2. Attempt to download content with a trusted but expired cert.
        3. Assert a 403 Unauthorized is returned.
        """
        with self.assertRaises(HTTPError) as raised_exception:
            set_distribution_base_path_and_download_a_content_unit_with_cert(
                self.distribution.pulp_href,
                RHSM_V1_ONE_VAR_BASE_PATH,
                self.repo.pulp_href,
                RHSM_CLIENT_CERT_TRUSTED_BUT_EXPIRED
            )
        self.assertEqual(raised_exception.exception.response.status_code, 403)


class RHSMV3CABundleCertGuardTestCase(RHSMCABundleCertGuardBase):
    """Api tests for RHSMCertGard with V3 RHSM Certificates and a bundle of CAs in a file."""

    def test_allow_request_when_cert_matches_zero_var_path(self):
        """
        Assert a correctly configured client can fetch content from a zero-variable path.

        1. Configure the distribution with a zero-variable path in the RHSM Cert.
        2. Attempt to download content.
        """
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            RHSM_V3_ZERO_VAR_BASE_PATH,
            self.repo.pulp_href,
            RHSM_V3_ZERO_VAR_CLIENT_CERT
        )



from logging import getLogger
from gettext import gettext as _
import re
from urllib.parse import unquote

from django.conf import settings
from django.db import models

from OpenSSL import crypto as openssl

from pulpcore.plugin.models import ContentGuard

from pulp_certguard.app.utils import get_rhsm

try:
    from rhsm import certificate
except ImportError:
    pass


logger = getLogger(__name__)

cert_unquoted_body_regex = re.compile('^-----BEGIN CERTIFICATE-----(.*)-----END CERTIFICATE-----')


class BaseCertGuard(ContentGuard):
    """A Base class all CertGuard implementations should derive from."""

    ca_certificate = models.TextField()

    @staticmethod
    def _reassemble_client_cert(unquoted_client_cert):
        match_result = cert_unquoted_body_regex.match(unquoted_client_cert)
        if match_result:
            cert_body = match_result.groups()[0]
            logger.debug("Reassembled client certificate")
            reassembled_client_cert = '-----BEGIN CERTIFICATE-----' + \
                                      cert_body.replace(' ', '\n') + \
                                      '-----END CERTIFICATE-----' + '\n'
            return reassembled_client_cert
        else:
            logger.debug("Did *not* reassemble client cert")
            return unquoted_client_cert

    @classmethod
    def _get_client_cert_header(cls, request):
        try:
            client_cert_data = request.headers["X-CLIENT-CERT"]
            logger.debug(f"client_cert_data received: {client_cert_data}")
        except KeyError:
            msg = _("A client certificate was not received via the `X-CLIENT-CERT` header.")
            raise PermissionError(msg)
        unquoted_client_cert = unquote(client_cert_data)
        return cls._reassemble_client_cert(unquoted_client_cert)

    def _ensure_client_cert_is_trusted(self, unquoted_certificate):
        trust_store = self._build_trust_store()

        try:
            openssl_client_cert = openssl.load_certificate(
                openssl.FILETYPE_PEM, buffer=unquoted_certificate
            )
        except openssl.Error as exc:
            raise PermissionError(str(exc))

        try:
            context = openssl.X509StoreContext(
                certificate=openssl_client_cert,
                store=trust_store,
            )
            context.verify_certificate()
        except openssl.X509StoreContextError as exc:
            if exc.args[0][0] == 20:  # The error code for client cert not signed by the CA
                msg = _("Client certificate is not signed by the stored 'ca_certificate'.")
                raise PermissionError(msg)
            if exc.args[0][0] == 10:  # The error code for an expired certificate
                msg = _("Client certificate is expired.")
                raise PermissionError(msg)
            raise PermissionError(str(exc))
        except openssl.Error as exc:
            raise PermissionError(str(exc))

    def _build_trust_store(self):
        trust_store = openssl.X509Store()

        # self.ca_certificate can be a **bundle** of certificates, which must be added to
        # X509Store one at a time.
        # We break ca_certificate up along BEGIN-CERTIFICATE/END-CERTIFICATE lines.
        # Certs can come with non-cert content (e.g. embedded-human-readable format inline);
        # hence, the following reg-ex throws out lines that don't exist between BEGIN/END pairs,
        # and builds certs out of the retained lines.
        rx = re.compile(
            r'(-----BEGIN CERTIFICATE-----.*?-----END CERTIFICATE-----)',
            re.MULTILINE | re.DOTALL
        )
        ca_certs = rx.findall(str(self.ca_certificate))

        # Now load each resulting cert, and set up in the trust_store
        for crt in ca_certs:
            try:
                openssl_ca_cert = openssl.load_certificate(
                    openssl.FILETYPE_PEM, buffer=crt.encode()
                )
            except openssl.Error as exc:
                raise PermissionError(str(exc))
            trust_store.add_cert(openssl_ca_cert)

        # If we get this far - return the now-loaded store!
        return trust_store

    class Meta:
        abstract = True


class RHSMCertGuard(BaseCertGuard):
    """
    A content-guard validating on a RHSM Certificate validated by `python-rhsm`.

    A Certificate Authority certificate to trust is required with each RHSMCertGuard created. With
    each request, the client certificate is first checked if it is signed by this CA cert. If not,
    it's untrusted and denied regardless of its paths.

    After determining the client certificate is trusted, the requested path is checked against the
    named paths in the certificate. A request is permitted if the current request path is a prefix
    of a path declared in the trusted RHSM Client Certificate.

    Fields:
        rhsm_certificate (models.TextField): The RHSM Certificate used to validate the client
            certificate at request time.
    """

    TYPE = 'rhsm'

    def __init__(self, *args, **kwargs):
        """Initialize a RHSMCertGuard and ensure this system has python-rhsm on it."""
        get_rhsm()  # Validate that rhsm is installed
        super().__init__(*args, **kwargs)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"

    def permit(self, request):
        """
        Validate the client cert is trusted and asserts a path that is prefix of the requested path.

        Args:
            request: The request from the user.

        Raises:
            PermissionError: If the request path is not a subpath of a path named in the
                certificate, or if the client certificate is not trusted from the CA certificated
                stored as `ca_certificate`.
        """
        get_rhsm()
        unquoted_certificate = self._get_client_cert_header(request)
        self._ensure_client_cert_is_trusted(unquoted_certificate)
        rhsm_cert = self._create_rhsm_cert_from_pem(unquoted_certificate)
        content_path_prefix_without_trail_slash = settings.CONTENT_PATH_PREFIX.rstrip('/')
        len_prefix_to_remove = len(content_path_prefix_without_trail_slash)
        path_without_content_path_prefix = request.path[len_prefix_to_remove:]
        self._check_paths(rhsm_cert, path_without_content_path_prefix)

    @staticmethod
    def _create_rhsm_cert_from_pem(unquoted_certificate):
        try:
            rhsm_cert = certificate.create_from_pem(unquoted_certificate)
        except certificate.CertificateException:
            msg = _("An error occurred while loading the client certificate data into python-rhsm.")
            raise PermissionError(msg)
        return rhsm_cert

    @staticmethod
    def _check_paths(rhsm_cert, path):
        logger.debug(f"Checking that path {path} is allowed in client cert")
        if rhsm_cert.check_path(path) is False:
            logger.debug(f"Path {path} is *not* allowed in client cert")
            msg = _("Requested path is not a subpath of a path in the client certificate.")
            raise PermissionError(msg)


class X509CertGuard(BaseCertGuard):
    """
    A content-guard that authenticates the request based on a client provided X.509 Certificate.

    Fields:
        ca_certificate (models.FileField): The CA certificate used to
            validate the client certificate.
    """

    TYPE = 'x509'

    def permit(self, request):
        """
        Validate the client cert is trusted.

        Args:
            request: The request from the user.

        Raises:
            PermissionError: If the client certificate is not trusted from the CA certificated
                stored as `ca_certificate`.
        """
        unquoted_certificate = self._get_client_cert_header(request)
        self._ensure_client_cert_is_trusted(unquoted_certificate)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"



# Generated by Django 2.2.11 on 2020-04-01 20:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('core', '0026_task_group'),
    ]

    operations = [
        migrations.CreateModel(
            name='RHSMCertGuard',
            fields=[
                ('contentguard_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='certguard_rhsmcertguard', serialize=False, to='core.ContentGuard')),
                ('ca_certificate', models.TextField()),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.contentguard',),
        ),
        migrations.CreateModel(
            name='X509CertGuard',
            fields=[
                ('contentguard_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='certguard_x509certguard', serialize=False, to='core.ContentGuard')),
                ('ca_certificate', models.TextField()),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.contentguard',),
        ),
    ]



"""Constants for Pulp certguard plugin tests."""
import os


_CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))


X509_BASE_PATH = "my-content-view"
X509_CERTS_BASE_PATH = os.path.join(
    _CURRENT_DIR, 'artifacts', 'x509', 'certificates'
)
X509_CA_CERT_FILE_PATH = os.path.join(X509_CERTS_BASE_PATH, 'ca.pem')
X509_CLIENT_CERT_FILE_PATH = os.path.join(X509_CERTS_BASE_PATH, 'client.pem')
X509_UNTRUSTED_CLIENT_CERT_FILE_PATH = os.path.join(X509_CERTS_BASE_PATH, 'untrusted_client.pem')
X509_UN_URLENCODED_CLIENT_CERT_FILE_PATH = os.path.join(
    X509_CERTS_BASE_PATH, 'un_urlencoded_cert.txt'
)


RHSM_CA_CERT_FILE_PATH = os.path.join(
    _CURRENT_DIR, 'artifacts', 'rhsm', 'katello-default-ca.crt'
)

RHSM_CLIENT_CERT_FROM_UNTRUSTED_CA = os.path.join(
    _CURRENT_DIR, 'artifacts', 'rhsm', 'untrusted_cert.pem'
)

RHSM_CLIENT_CERT_TRUSTED_BUT_EXPIRED = os.path.join(
    _CURRENT_DIR, 'artifacts', 'rhsm', 'trusted_but_expired.pem'
)

THIRDPARTY_CA_CERT_FILE_PATH = os.path.join(
    _CURRENT_DIR, 'artifacts', 'thirdparty_ca', 'certificates', 'ca.pem'
)


# Uber cert path: /Default_Organization
RHSM_UBER_CLIENT_CERT = os.path.join(
    _CURRENT_DIR, 'artifacts', 'rhsm', 'uber.cert'
)
RHSM_UBER_CERT_BASE_PATH_ONE = "Default_Organization/my-content-view"
RHSM_UBER_CERT_BASE_PATH_TWO = "Default_Organization/another-content-view"


# Zero_var path: /Default_Organization/Library/custom/foo/foo
RHSM_V1_ZERO_VAR_CLIENT_CERT = os.path.join(
    _CURRENT_DIR, 'artifacts', 'rhsm', 'v1', '159442575569388840.pem'
)
RHSM_V1_ZERO_VAR_BASE_PATH = "Default_Organization/Library/custom/foo/foo"


# One var path: /Default_Organization/Library/content/dist/rhel/server/7/7Server/$basearch/extras/os
# Two var path: /Default_Organization/Library/content/dist/rhel/server/7/$releasever/$basearch/os
RHSM_V1_ONE_AND_TWO_VAR_CLIENT_CERT = os.path.join(
    _CURRENT_DIR, 'artifacts', 'rhsm', 'v1', '1514454871848760713.pem'
)
RHSM_V1_ONE_VAR_BASE_PATH = \
    "Default_Organization/Library/content/dist/rhel/server/7/7Server/x86_64/extras/os"
RHSM_V1_TWO_VAR_BASE_PATH = "Default_Organization/Library/content/dist/rhel/server/7/7.4/x86_64/os"


# Zero_var path: /Default_Organization/Library/custom/foo/foo
RHSM_V3_ZERO_VAR_CLIENT_CERT = os.path.join(
    _CURRENT_DIR, 'artifacts', 'rhsm', 'v3', '4260035510644027985.pem'
)
RHSM_V3_ZERO_VAR_BASE_PATH = "Default_Organization/Library/custom/foo/foo"
RHSM_V3_INVALID_BASE_PATH = "this-is-not-a-valid-base-path"


# One var path: /Default_Organization/Library/content/dist/rhel8/$releasever/x86_64/baseos/os
# Two var path: /Default_Organization/Library/content/dist/rhel/server/7/$releasever/$basearch/os
RHSM_V3_ONE_AND_TWO_VAR_CLIENT_CERT = os.path.join(
    _CURRENT_DIR, 'artifacts', 'rhsm', 'v3', '5527980418107729172.pem'
)
RHSM_V3_ONE_VAR_BASE_PATH = "Default_Organization/Library/content/dist/rhel8/8/x86_64/baseos/os"
RHSM_V3_TWO_VAR_BASE_PATH = "Default_Organization/Library/content/dist/rhel/server/7/7.4/x86_64/os"



from gettext import gettext as _

from django.core.exceptions import ImproperlyConfigured

try:
    import rhsm  # rhsm is an optional dependency
except ImportError as e:
    rhsm = None
    rhsm_import_error = str(e)


def get_rhsm():
    """
    Returns the `rhsm` module or raises an exception if `rhsm` is not installed.

    `rhsm` is an optional dependency so you can call this function to get `rhsm` or have an
    `ImproperlyConfigured` error for the user.

    Returns:
        The `rhsm` Python module.

    Raises:
        `ImproperlyConfigured` exception explaining `rhsm` is not installed.
    """
    if rhsm is None:
        error_msg = _("RHSMCertGuard requires the Python package 'rhsm' to be installed ({}).")
        raise ImproperlyConfigured(error_msg.format(rhsm_import_error))
    return rhsm



"""Utilities for tests for the certguard plugin."""

from random import choice
from time import sleep
from urllib.parse import quote

from pulp_smash import config
from pulp_smash.pulp3.utils import download_content_unit

from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    TasksApi,
)

from pulpcore.client.pulp_certguard import ApiClient as CertguardApiClient

from pulpcore.client.pulp_file import (
    DistributionsFileApi,
    RepositoriesFileApi,
)

from pulp_file.tests.functional.utils import (
    gen_file_client,
    get_file_content_paths,
)


cfg = config.get_config()
configuration = cfg.get_bindings_config()

core_client = CoreApiClient(configuration)
tasks = TasksApi(core_client)


def read_cert(cert_path):
    """Return an string of data read from `cert_path`."""
    with open(cert_path, 'r') as cert_file:
        return cert_file.read()


def read_cert_and_urlencode(cert_path):
    """Return an string of data read from `cert_path` and urlencode."""
    return quote(read_cert(cert_path))


def gen_certguard_client():
    """Return an OBJECT for file client."""
    return CertguardApiClient(configuration)


def monitor_task(task_href):
    """Polls the Task API until the task is in a completed state.

    Prints the task details and a success or failure message. Exits on failure.

    Args:
        task_href(str): The href of the task to monitor

    Returns:
        list[str]: List of hrefs that identify resource created by the task

    """
    completed = ["completed", "failed", "canceled"]
    task = tasks.read(task_href)
    while task.state not in completed:
        sleep(2)
        task = tasks.read(task_href)

    if task.state == "completed":
        return task.created_resources

    return task.to_dict()


def set_distribution_base_path(file_distribution_href, base_path):
    """
    Set the base path of a FileDistribution and return the updated representation.

    Args:
        file_distribution_href: The distribution href that is to be updated. This must refer to a
            distribution of type `FileDistribution`.
        base_path: The base path to set on the `distribution`.

    Returns:
        The bindings object representing the updated FileDistribution.
    """
    file_client = gen_file_client()
    distributions_api = DistributionsFileApi(file_client)
    update_response = distributions_api.partial_update(
        file_distribution_href,
        {"base_path": base_path}
    )
    monitor_task(update_response.task)
    return distributions_api.read(file_distribution_href)


def set_distribution_base_path_and_download_a_content_unit_with_cert(
        file_distribution_href,
        base_path,
        file_repository_href,
        cert_path,
        content_path=None,
        url_encode=True):
    """
    Set the base path on the `distribution, read the cert, urlencode it, and then request one unit.

    If `content_path` is set, that path will be requested, otherwise a random, valid content unit
    path will be selected from the FileRepository at `file_repository_href`.

    1. Set the distribution referred to by `file_distribution_href` base_path to `base_path`.
    2. Read the cert from the filesystem and urlencode it.
    3. Make a request to `content_path` if specified, or to a random content item present in the
        `file_repository_href` repository. The urlencoded cert is submitted as the `X-CLIENT-CERT`
        header when requesting content.

    Args:
        file_distribution_href: The distribution href that is to be updated. This must refer to a
            distribution of type `FileDistribution`.
        base_path: The base path to set on the `distribution`.
        file_repository_href: The repository href that will have
        cert_path: The file system path to the certificate to be used in the content request. This
            will be read from the filesystem and urlencoded before being submitted as the
            `X-CLIENT-CERT` header when downloading content.
        content_path: The path to the specific content unit to be fetched. This is the portion of
            the url after the distribution URL. It's optional, and if unspecified a random, valid
            content unit will be selected instead from the repository.
        url_encode: If true, the certificate data read will be urlencoded, otherwise it won't be.
            This is an optional param, and defaults to True.

    Returns:
        The downloaded data.

    """
    distribution = set_distribution_base_path(file_distribution_href, base_path)

    if content_path is None:
        file_client = gen_file_client()
        file_repos_api = RepositoriesFileApi(file_client)
        repo = file_repos_api.read(file_repository_href)
        content_path = choice(get_file_content_paths(repo.to_dict()))

    if url_encode:
        cert_data = read_cert_and_urlencode(cert_path)
    else:
        cert_data = read_cert(cert_path)

    return download_content_unit(
        config.get_config(),
        distribution.to_dict(),
        content_path,
        headers={'X-CLIENT-CERT': cert_data}
    )



from gettext import gettext as _

from OpenSSL import crypto as openssl

from pulpcore.plugin.serializers import ContentGuardSerializer

from rest_framework import serializers

from pulp_certguard.app.utils import get_rhsm
from .models import RHSMCertGuard, X509CertGuard


class BaseCertGuardSerializer(ContentGuardSerializer):
    """A Base Serializer class for all Cert Guard Serializers."""

    ca_certificate = serializers.CharField(
        help_text=_(
            "A Certificate Authority (CA) certificate (or a bundle thereof) "
            "used to verify client-certificate authenticity."
        ),
    )

    @staticmethod
    def validate_ca_certificate(ca_certificate):
        """Validates the given certificate as a PEM encoded X.509 certificate using openssl."""
        try:
            openssl.load_certificate(openssl.FILETYPE_PEM, buffer=ca_certificate)
        except ValueError:
            reason = _('Must be PEM encoded X.509 certificate.')
            raise serializers.ValidationError(reason)
        else:
            return ca_certificate

    class Meta:
        fields = ContentGuardSerializer.Meta.fields + (
            'ca_certificate',
        )


class RHSMCertGuardSerializer(BaseCertGuardSerializer):
    """RHSM Content Guard Serializer."""

    class Meta:
        model = RHSMCertGuard
        fields = BaseCertGuardSerializer.Meta.fields

    @staticmethod
    def validate_ca_certificate(ca_certificate):
        """Validates the given certificate."""
        get_rhsm()  # Validate that rhsm is installed
        return BaseCertGuardSerializer.validate_ca_certificate(ca_certificate)


class X509CertGuardSerializer(BaseCertGuardSerializer):
    """X.509 Content Guard Serializer."""

    class Meta:
        model = X509CertGuard
        fields = BaseCertGuardSerializer.Meta.fields



from pulpcore.plugin.viewsets import ContentGuardFilter, ContentGuardViewSet

from .models import RHSMCertGuard, X509CertGuard
from .serializers import RHSMCertGuardSerializer, X509CertGuardSerializer


class RHSMCertGuardViewSet(ContentGuardViewSet):
    """RHSMCertGuard API Viewsets."""

    endpoint_name = 'rhsm'
    queryset = RHSMCertGuard.objects.all()
    serializer_class = RHSMCertGuardSerializer
    filterset_class = ContentGuardFilter


class X509CertGuardViewSet(ContentGuardViewSet):
    """X509CertGuard API Viewsets."""

    endpoint_name = 'x509'
    queryset = X509CertGuard.objects.all()
    serializer_class = X509CertGuardSerializer
    filterset_class = ContentGuardFilter



import uuid

from pulpcore.client.pulp_certguard import CertguardX509CertGuard, ContentguardsX509Api

from pulp_certguard.tests.functional.api.base import BaseCertGuard, CommonDenialTestsMixin
from pulp_certguard.tests.functional.constants import (
    X509_BASE_PATH,
    X509_CA_CERT_FILE_PATH,
    X509_CLIENT_CERT_FILE_PATH,
    X509_UNTRUSTED_CLIENT_CERT_FILE_PATH,
    X509_UN_URLENCODED_CLIENT_CERT_FILE_PATH,
)
from pulp_certguard.tests.functional.utils import (
    gen_certguard_client,
    set_distribution_base_path_and_download_a_content_unit_with_cert,
)


class X509CertGuardTestCase(BaseCertGuard, CommonDenialTestsMixin):
    """Api tests for X509CertGard."""

    DENIALS_BASE_PATH = X509_BASE_PATH
    UNTRUSTED_CLIENT_CERT_PATH = X509_UNTRUSTED_CLIENT_CERT_FILE_PATH

    @classmethod
    def _setup_content_guard(cls):
        # Create an X.509 Content Guard
        certguard_client = gen_certguard_client()
        cls.x509_content_guards_api = ContentguardsX509Api(certguard_client)

        with open(X509_CA_CERT_FILE_PATH, 'r') as x509_ca_cert_data_file:
            x509_ca_cert_data = x509_ca_cert_data_file.read()

        x509_cert_guard = CertguardX509CertGuard(
            name=str(uuid.uuid4()),
            ca_certificate=x509_ca_cert_data
        )
        cls.x509_content_guard_data = cls.x509_content_guards_api.create(x509_cert_guard)
        cls.teardown_cleanups.append(
            (cls.x509_content_guards_api.delete, cls.x509_content_guard_data.pulp_href)
        )
        return cls.x509_content_guard_data.pulp_href

    def test_allow_request_when_cert_is_trusted(self):
        """
        Assert a correctly configured client can fetch content.

        1. Configure the distribution with an X.509 CertGuard.
        2. Attempt to download content.
        """
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            X509_BASE_PATH,
            self.repo.pulp_href,
            X509_CLIENT_CERT_FILE_PATH
        )

    def test_allow_request_when_apache_un_urlencoded_cert_is_trusted(self):
        """
        Assert a correctly configured client can fetch content with reverse proxy Apache < 2.6.10.

        1. Configure the distribution with an X.509 CertGuard.
        2. Attempt to download content with an un-urlencoded certificate (Apache < 2.6.10 style)
        """
        set_distribution_base_path_and_download_a_content_unit_with_cert(
            self.distribution.pulp_href,
            X509_BASE_PATH,
            self.repo.pulp_href,
            X509_UN_URLENCODED_CLIENT_CERT_FILE_PATH,
            url_encode=False,
        )



from requests import HTTPError
import unittest

from pulp_smash import config

from pulpcore.client.pulp_file import (
    DistributionsFileApi,
    FileFilePublication,
    PublicationsFileApi,
    RepositoriesFileApi,
    RepositorySyncURL,
    RemotesFileApi,
)

from pulp_file.tests.functional.utils import (
    gen_file_client,
    gen_file_remote,
)
from pulp_smash.pulp3.utils import (
    download_content_unit,
    gen_distribution,
    gen_repo,
)

from pulp_certguard.tests.functional.utils import (
    monitor_task,
    set_distribution_base_path,
    set_distribution_base_path_and_download_a_content_unit_with_cert,
)


class BaseCertGuard(unittest.TestCase):
    """A base class for all CertGuard tests which provides generic setup and teardown."""

    @classmethod
    def setUpClass(cls):
        """
        Initialize Pulp to make authorization assertions using client certificates.

        0. Create a FileRepository
        1. Create a FileRemote
        2. Sync in a few units we can use to fetch with
        3. Create a Publication
        4. Create a CertGuard with the CA cert used to sign all client certificates
        5. Create a Distribution for the publication that is protected by the CertGuard

        """
        cls.teardown_cleanups = []
        cls.cfg = config.get_config()

        file_client = gen_file_client()
        repo_api = RepositoriesFileApi(file_client)
        remote_api = RemotesFileApi(file_client)
        publications = PublicationsFileApi(file_client)
        cls.distributions_api = DistributionsFileApi(file_client)

        cls.repo = repo_api.create(gen_repo())
        cls.teardown_cleanups.append((repo_api.delete, cls.repo.pulp_href))

        body = gen_file_remote(policy="immediate")
        remote = remote_api.create(body)
        cls.teardown_cleanups.append((remote_api.delete, remote.pulp_href))

        # Sync a Repository
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(cls.repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        cls.repo = repo_api.read(cls.repo.pulp_href)

        # Create a publication.
        publish_data = FileFilePublication(repository=cls.repo.pulp_href)
        publish_response = publications.create(publish_data)
        created_resources = monitor_task(publish_response.task)
        publication_href = created_resources[0]
        cls.teardown_cleanups.append((publications.delete, publication_href))

        content_guard_href = cls._setup_content_guard()

        # Create a distribution.
        body = gen_distribution()
        body["publication"] = publication_href
        body["content_guard"] = content_guard_href
        distribution_response = cls.distributions_api.create(body)
        created_resources = monitor_task(distribution_response.task)
        cls.distribution = cls.distributions_api.read(created_resources[0])
        cls.teardown_cleanups.append((cls.distributions_api.delete, cls.distribution.pulp_href))

    @classmethod
    def _setup_content_guard(cls):
        raise NotImplementedError("subclasses must implement '_setup_content_guard'")

    @classmethod
    def tearDownClass(cls):
        """Clean class-wide variable."""
        for cleanup_function, args in reversed(cls.teardown_cleanups):
            cleanup_function(args)


class CommonDenialTestsMixin:
    """Common tests between X.509 and RHSM Cert Guards which should all produce denials."""

    DENIALS_BASE_PATH = None
    UNTRUSTED_CLIENT_CERT_PATH = None

    def test_denial_when_empty_client_header_submitted(self):
        """
        Assert denial when a client submits an empty X-CLIENT-CERT header.

        1. Configure the distribution with a valid base path.
        2. Attempt to download content.
        3. Assert a 403 Unauthorized is returned.
        """
        distribution = set_distribution_base_path(
            self.distribution.pulp_href,
            self.DENIALS_BASE_PATH
        )

        content_path = ""

        with self.assertRaises(HTTPError) as raised_exception:
            download_content_unit(
                config.get_config(),
                distribution.to_dict(),
                content_path,
                headers={'X-CLIENT-CERT': ""}
            )
        self.assertEqual(raised_exception.exception.response.status_code, 403)

    def test_denial_when_no_client_header_submitted(self):
        """
        Assert denial when a client submits no X-CLIENT-CERT header.

        1. Configure the distribution with a valid base path.
        2. Attempt to download content.
        3. Assert a 403 Unauthorized is returned.
        """
        distribution = set_distribution_base_path(
            self.distribution.pulp_href,
            self.DENIALS_BASE_PATH
        )

        content_path = ""

        with self.assertRaises(HTTPError) as raised_exception:
            download_content_unit(
                config.get_config(),
                distribution.to_dict(),
                content_path
            )
        self.assertEqual(raised_exception.exception.response.status_code, 403)

    def test_denial_when_client_header_contains_an_invalid_certificate(self):
        """
        Assert denial when a client submits X-CLIENT-CERT header with invalid certificate data.

        1. Configure the distribution with a valid base path.
        2. Attempt to download content.
        3. Assert a 403 Unauthorized is returned.
        """
        distribution = set_distribution_base_path(
            self.distribution.pulp_href,
            self.DENIALS_BASE_PATH
        )

        content_path = ""

        with self.assertRaises(HTTPError) as raised_exception:
            download_content_unit(
                config.get_config(),
                distribution.to_dict(),
                content_path,
                headers={'X-CLIENT-CERT': "this is not cert data"}
            )
        self.assertEqual(raised_exception.exception.response.status_code, 403)

    def test_denial_when_client_header_contains_an_untrusted_certificate(self):
        """
        Assert denial when a client submits a valid but rhsm certificate but not for the trusted CA.

        1. Configure the distribution with a valid base path.
        2. Attempt to download content with an untrusted client certificate.
        3. Assert a 403 Unauthorized is returned.
        """
        with self.assertRaises(HTTPError) as raised_exception:
            set_distribution_base_path_and_download_a_content_unit_with_cert(
                self.distribution.pulp_href,
                self.DENIALS_BASE_PATH,
                self.repo.pulp_href,
                self.UNTRUSTED_CLIENT_CERT_PATH
            )
        self.assertEqual(raised_exception.exception.response.status_code, 403)



#!/usr/bin/python
#
# This code is contribued to and licencensed under the terms of the
# pulp-certguard project as a whole.
#
# The original author officially turns over all licensing rights to
# the pulp project/maintainers.


from contextlib import closing
from yum.plugins import TYPE_CORE

requires_api_version = '2.3'
plugin_type = (TYPE_CORE,)

ID_SSL_HEADER = "X-CLIENT-CERT"
CERT_PATH = "/etc/boomi/yum.pem"
# Repos that begin with this prefix will be handled by this plugin
REPO_PREFIX = "boomi-"


def init_hook(conduit):
    """Plugin initialization hook.

    For each boomi repo, replace yum's representation of the repo with
    a subclass that adds in the necessary headers.

    """
    # Only process boomi repos
    repos = conduit.getRepos()
    boomi_repos = repos.findRepos(REPO_PREFIX)

    # Retrieve the Amazon metadata
    id_cert = _load_cert()

    if id_cert:
        # Add the headers to all BOOMI repos
        for repo in boomi_repos:
            repo.http_headers[ID_SSL_HEADER] = id_cert


def _load_cert():
    """Loads and returns the certificate.

    @rtype: string
    """
    with(closing(open(CERT_PATH, 'r'))) as fd:
        id_doc = fd.read()
    # Newlines in http headers aren't valid
    return id_doc.replace("\n", "")



from pulpcore.plugin import PulpPluginAppConfig


class PulpCertGuardPluginAppConfig(PulpPluginAppConfig):
    """App config for cert guard plugin."""

    name = 'pulp_certguard.app'
    label = 'certguard'
    version = '1.5.0'
